import sys
import math
import random
import pygame

import data
import const

class GameWindow(object):
    
    def __init__(self):
        pygame.display.set_caption('Beam')
        self.screen = pygame.display.set_mode((const.WIDTH, const.HEIGHT))
        try:
            pygame.mixer.init()
            pygame.mixer.music.load(data.filepath('Rhinoceros.mp3'))
            pygame.mixer.music.play(-1)
        except:
            print 'Cannot load music.'
        ret = 0
        while ret == 0:
            self.intro()
            ret = self.game(1)
            while ret > 0:
                ret = self.game(ret)
        
    def intro(self):
        Intro(self, self.screen, True).loop()
        Intro(self, self.screen, False).loop()
    
    def game(self, level):
        return Game(self, self.screen, level).loop()


class Intro(object):

    def __init__(self, window, screen, first):
        self.window = window
        self.screen = screen
        self.first = first

        self.muted = False

    def loop(self):
        if self.first:
            self.bg = pygame.image.load(data.filepath('intro1.png')).convert()
        else:
            self.bg = pygame.image.load(data.filepath('intro2.png')).convert()
        self.screen.blit(self.bg, (0,0))
        pygame.display.update()

        self.done = False
        while not self.done:
            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    sys.exit()
                elif event.type == pygame.KEYDOWN:
                    if event.key == pygame.K_m:
                        if not self.muted:
                            pygame.mixer.music.pause()
                            self.muted = True
                        else:
                            pygame.mixer.music.unpause()
                            self.muted = False
                    elif event.key == pygame.K_ESCAPE:
                        sys.exit()
                    else:
                        self.done = True

            pygame.display.flip()


class Game(object):
    
    def __init__(self, window, screen, level):
        self.window = window
        self.screen = screen
        self.level = level
        self.clock = pygame.time.Clock()
        
        self.paused = False
        self.muted = False
    
    def loop(self):
        self.max_level = 10
        if self.level > self.max_level: return 0
        self.bg = pygame.image.load(data.filepath('bg' + str(self.level) + '.png')).convert()
        if self.level < self.max_level:
            self.bg_floor = pygame.image.load(data.filepath('bg_floor.png')).convert_alpha()        
            self.bg.blit(self.bg_floor, (0,0))
        self.screen.blit(self.bg, (0,0))
        pygame.display.update()
        
        self.player = Player((-32,const.HEIGHT/2))
        self.player_con = pygame.sprite.RenderUpdates(self.player)
        
        self.laser_mgr = LaserManager(self.player, self.level)
        
        self.roomlaser = RoomLaser((-180,0), self.player)
        self.roomlaser_con = pygame.sprite.RenderUpdates(self.roomlaser)

        if self.level == self.max_level:
            self.roomlaser.start = False
        
        self.keys = [pygame.K_z, pygame.K_x, pygame.K_SPACE]
        for k in self.keys:
            if pygame.key.get_pressed()[k]: self.player.stop_shield()

        self.done = False
        while not self.done:
            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    sys.exit()
                elif event.type == pygame.KEYDOWN:
                    self.on_keydown(event)
                elif event.type == pygame.KEYUP:
                    self.on_keyup(event)

            # clear
            if not self.paused:
                self.laser_mgr.body_con.clear(self.screen, self.bg)
                self.laser_mgr.clear(self.screen, self.bg)
                self.player_con.clear(self.screen, self.bg)
                self.roomlaser_con.clear(self.screen, self.bg)

            dt = self.clock.tick(const.FPS) / 1000.0

            # update
            if not self.paused:
                self.laser_mgr.update(dt)
                self.player.update(dt)
                if self.player.win: self.next_level()
                self.roomlaser.update(dt)

            # draw
            if not self.paused:
                self.laser_mgr.body_con.draw(self.screen)
                self.laser_mgr.draw(self.screen)
                self.player_con.draw(self.screen)
                self.roomlaser_con.draw(self.screen)

            pygame.display.flip()
        
        return self.level + 1

    def pause(self):
        self.paused = not self.paused

    def on_keydown(self, event):
        if event.key == pygame.K_ESCAPE:
            self.done = True
            self.level = -1
        elif event.key == pygame.K_p:
            self.pause()
        elif event.key == pygame.K_r:
            self.done = True
            self.level = 0
        elif event.key == pygame.K_m:
            if not self.muted:
                pygame.mixer.music.pause()
                self.muted = True
            else:
                pygame.mixer.music.unpause()
                self.muted = False
        elif event.key in self.keys:
            self.player.stop_shield()
            if self.player.dead:
                self.done = True
                self.level = 0

    def on_keyup(self, event):
        if event.key in self.keys:
            self.player.start_shield()
    
    def next_level(self):
        self.done = True    


class Player(pygame.sprite.Sprite):
    
    def __init__(self, pos):
        pygame.sprite.Sprite.__init__(self)
        self.pos = list(pos)
        self.image_run = pygame.image.load(data.filepath('player.png')).convert_alpha()
        self.image_shield = pygame.image.load(data.filepath('player_shield.png')).convert_alpha()
        self.image_dead = pygame.image.load(data.filepath('player_dead.png')).convert_alpha()
        self.image = self.image_shield
        self.rect = self.image.get_rect()
        self.rect.center = self.pos
        self.speed = const.SHIELD_SPEED
        self.vel = [0, 0]
        self.moving_left, self.moving_right, self.moving_up, self.moving_down = False, False, False, False
        self.shield = False
        self.dead = False
        self.win = False
        self.laser_sound = pygame.mixer.Sound(data.filepath('laser.wav'))
        self.laser_sound.set_volume(0.5)
    
    def start_shield(self): 
        if self.dead: return
        self.shield = True
        self.speed = const.SHIELD_SPEED
        self.image = self.image_shield
        self.rect = self.image.get_rect()
        self.rect.center = self.pos

    def stop_shield(self): 
        if self.dead: return
        self.shield = False
        self.speed = const.PLAYER_SPEED
        self.image = self.image_run
        self.rect = self.image.get_rect()
        self.rect.center = self.pos

    def update(self, dt):
        self.pos[0] += self.speed * dt
        if self.pos[0] <= 0:
            self.pos[0] = 0
        if self.pos[0] + self.rect.width >= const.WIDTH+20:
            self.pos[0] = const.WIDTH+20 - self.rect.width
            self.win = True
        self.rect.center = self.pos
    
    def die(self):
        if not self.dead:
            self.image = self.image_dead
            self.rect = self.image.get_rect()
            self.rect.center = self.pos
            self.speed = 0
            self.dead = True
            self.laser_sound.play(0)


class LaserManager(pygame.sprite.OrderedUpdates):

    def __init__(self, player, level):
        pygame.sprite.OrderedUpdates.__init__(self)
        self.tilesize = 64
        self.sizex = 8
        self.sizey = 8
        self.pos = (10,const.HEIGHT / 2)
        self.player = player
        
        self.body_con = pygame.sprite.RenderUpdates()

        self.ne_lasers = []
        self.se_lasers = []
        self.ne_lasers_tiles = []
        self.se_lasers_tiles = []        
        self.ne_lasers_on = []
        self.se_lasers_on = []
        
        self.level = level
        lasers = [1,2,2,3,3,3,4,4,4,0]
        for i in range(lasers[self.level-1]):
            self.generate_laser()
    
    def generate_laser(self):
        direction = random.choice([0,1])
        if direction == 0:
            y = random.randint(1, self.sizey - 2)
            while y in self.ne_lasers:
                y = random.randint(1, self.sizey - 2)
            self.add_ne_laser(y)
        elif direction == 1:
            x = random.randint(1, self.sizex - 2)
            while x in self.se_lasers:
                x = random.randint(1, self.sizex - 2)
            self.add_se_laser(x)

    def add_ne_laser(self, y):
        if y in self.ne_lasers:
            return
        self.ne_lasers.append(y)

        chosen = []
        for i in range(random.choice(const.BODY_DIST)):
            body_x = random.randint(0, self.sizex - 1)
            while body_x in chosen:
                body_x = random.randint(0, self.sizex - 1)
            chosen.append(body_x)
            tpos = self.convert_to_rotated(body_x, y)
            tile = Body(tpos)
            self.body_con.add(tile)

        tile = LaserTile(self.convert_to_rotated(3.5, y), 0)
        self.add(tile)
        self.ne_lasers_tiles.append(tile)
        self.ne_lasers_on.append(False)

    def add_se_laser(self, x):
        if x in self.se_lasers:
            return
        self.se_lasers.append(x)

        chosen = []
        for i in range(random.choice(const.BODY_DIST)):
            body_y = random.randint(0, self.sizey - 1)
            while body_y in chosen:
                body_y = random.randint(0, self.sizey - 1)
            chosen.append(body_y)
            tpos = self.convert_to_rotated(x, body_y)
            tile = Body(tpos)
            self.body_con.add(tile)

        tile = LaserTile(self.convert_to_rotated(x, 3.5), 1)
        self.add(tile)
        self.se_lasers_tiles.append(tile)
        self.se_lasers_on.append(False)
    
    def update(self, dt):
        tol = 16
        for i, y in enumerate(self.ne_lasers):
            pos_x = self.convert_to_rotated(y,y)[0]
            if abs(self.player.rect.centerx - pos_x) < tol:
                if not self.ne_lasers_on[i]:
                    self.ne_lasers_on[i] = True
                    self.ne_lasers_tiles[i].set_visible()
            else:
                if self.ne_lasers_on[i]:
                    self.ne_lasers_on[i] = False
                    self.ne_lasers_tiles[i].set_invisible()
        for i, x in enumerate(self.se_lasers):
            pos_x = self.convert_to_rotated(x,x)[0]
            if abs(self.player.rect.centerx - pos_x) < tol:
                if not self.se_lasers_on[i]:
                    self.se_lasers_on[i] = True
                    self.se_lasers_tiles[i].set_visible()
            else:
                if self.se_lasers_on[i]:
                    self.se_lasers_on[i] = False
                    self.se_lasers_tiles[i].set_invisible()
        if not self.player.shield:
            for i, y in enumerate(self.ne_lasers):
                if self.ne_lasers_on[i]:
                    self.player.die()
            for i, x in enumerate(self.se_lasers):
                if self.se_lasers_on[i]:
                    self.player.die()

    def convert_to_rotated(self, x, y):
        rx = self.pos[0] + (y + x + 1) * self.tilesize * math.sqrt(2) * 0.5
        ry = self.pos[1] + (y - x) * self.tilesize * math.sqrt(2) * 0.5
        return (rx, ry)


class Body(pygame.sprite.Sprite):

    def __init__(self, pos):
        pygame.sprite.Sprite.__init__(self)
        self.pos = list(pos)
        self.image = pygame.image.load(data.filepath('body.png')).convert_alpha()
        self.rect = self.image.get_rect()
        self.rect.center = self.pos


class LaserTile(pygame.sprite.Sprite):

    def __init__(self, pos, direction):
        pygame.sprite.Sprite.__init__(self)
        self.pos = list(pos)
        self.dir = direction
        self.blank_image = pygame.Surface((1,1))
        self.blank_image.set_colorkey((0,0,0))
        if self.dir == 0:
            self.laser_image = pygame.image.load(data.filepath('laser_ne.png')).convert_alpha()
        elif self.dir == 1:
            self.laser_image = pygame.image.load(data.filepath('laser_se.png')).convert_alpha()
        self.image = self.blank_image
        self.rect = self.image.get_rect()
        self.rect.center = self.pos

    def set_visible(self):
        self.image = self.laser_image
        self.rect = self.image.get_rect()
        self.rect.center = self.pos

    def set_invisible(self):
        self.image = self.blank_image
        self.rect = self.image.get_rect()
        self.rect.center = self.pos


class RoomLaser(pygame.sprite.Sprite):
    
    def __init__(self, pos, player):
        pygame.sprite.Sprite.__init__(self)
        self.pos = list(pos)
        self.image = pygame.image.load(data.filepath('laser_room.png')).convert_alpha()
        self.rect = self.image.get_rect()
        self.rect.center = self.pos
        self.speed = const.LASER_SPEED
        self.start = True
        self.player = player
    
    def start_laser(self):
        self.start = True
    
    def update(self, dt):
        if not self.start: return
        self.pos[0] += self.speed * dt
        self.rect.topleft = self.pos
        if self.player.rect.centerx <= self.rect.x + 35:
            self.player.die()